-- ROM/RAM emulator
--

-- 
-- ANZX[hƊe[q̏Ԃ𐮗Ă
--                         (nEXCAS)(nRAS2)(_DRD2)
-- _MREQ _RD _WR _WE _CS2/3  _CAS   _RAS   _VRD

--   0    1   0   0    1      0      0     1    Ext-RAM Write(P6/mk2/SR)
--   0    0   1   1    1      0      0     0    Ext-RAM Read (P6/mk2/SR)
--   1    1   1   1    1      0      0     0    Ext-RAM Read from VDP (P6)

--   0    1   0   0    1      0      1     1    Int-RAM Write(P6/mk2)
--   0    1   0   0    1      1      0     1    Int-RAM Write(SR)
--   0    0   1   1    1      0      1     1    Int-RAM Read (P6/mk2)
--   0    0   1   1    1      1      0     1    Int-RAM Read (SR)
--   1    1   1   1    1      0      1     1    Int-RAM Read from VDP (P6)

--   0    0   1   1    0      1      1     1    Ext-ROM Read(P6/mk2)
--   0    0   1   1    0      1      0     1    Ext-ROM Read(SR)

--   0    0   1   1    1      1      1     1    Int-ROM Read(P6/mk2)
--   0    0   1   1    1      1      0     1    Int-ROM Read(SR)


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

entity ROMEMU is
	port (
		C_NRST			: in    std_logic;	-- from P6
		C_4MCLK			: in    std_logic;	-- from P6
		C_NWR			: in    std_logic;	-- from P6
		C_NRD			: in    std_logic;	-- from P6
		C_NIORQ			: in    std_logic;	-- from P6
		C_NMREQ			: in    std_logic;	-- from P6
		C_NM1			: in    std_logic;	-- from P6
		C_NEXCAS		: in    std_logic;	-- from P6
		C_NOCS5			: in    std_logic;	-- from P6
		C_NWE			: in    std_logic;	-- from P6
		C_NDRD2			: in    std_logic;	-- from P6
		C_NCS2			: in    std_logic;	-- from P6
		C_NRAS2			: in    std_logic;	-- from P6
		C_NCS3			: in    std_logic;	-- from P6
		C_NEXINT		: out   std_logic;	-- to P6
		C_NEROM			: out   std_logic;	-- to P6

		R0_NCS			: out   std_logic;	-- to SRAM (ROM emu)
		R1_NCS			: out   std_logic;	-- to SRAM
		R_NOE			: out   std_logic;	-- to SRAM
		R0_NWR			: out   std_logic;	-- to SRAM (ROM emu)
		R1_NWR			: out   std_logic;	-- to SRAM

		C_D				: inout std_logic_vector(7 downto 0);	-- Data bus
		C_A				: inout std_logic_vector(15 downto 0);	-- Address bus
		R_A				: out   std_logic_vector(18 downto 13);	-- Address bas for SRAM
		NADOUTENB		: out   std_logic;						-- Address/Data bus output enable
		NDOUTCNT		: out   std_logic;						-- Address bus output control

		PA				: inout std_logic_vector(7 downto 0);	-- AVR I/F
		PD				: in    std_logic_vector(6 downto 0);	-- AVR I/F
		NRSTOUT			: out   std_logic;						-- AVR I/F
		SW				: in    std_logic_vector(4 downto 1);	-- DIPSW

		CLK50M			: in    std_logic

	);
end ROMEMU;

architecture RTL of ROMEMU is

-- Memory Bank Register
	type Bank_type is array (0 to 7) of std_logic_vector(5 downto 0);
	signal bankset		: Bank_type;		-- AreaA-H

-- ROM/RAM set selector;
	signal romset		: std_logic_vector(1 downto 0);	--ROM setI (1set = 1M(16bank)) x 4set
	signal ramset		: std_logic_vector(1 downto 0);	--RAM setI (1set = 1M(16bank)) x 4set

	signal cs01r		: std_logic;					-- 0000-1FFFH Read  enable flag
	signal cs01rx		: std_logic;					-- 0000-1FFFH BASIC-ROM read
	signal cs01w		: std_logic;					-- 0000-1FFFH Write enable flag
	signal cs02w		: std_logic;					-- 2000-3FFFH Write enable flag
	signal pc60mode		: std_logic;					-- PC-6001łtO
	signal romxmode		: std_logic;					-- BASIC-ROM emu
	signal cscg			: std_logic;					-- CG read

	signal vdgread_n	: std_logic;
	signal read_ab_n	: std_logic;
	signal romemu_n		: std_logic;
	signal write_ab_n	: std_logic;
	signal write_cd_n	: std_logic;
	signal ramwrn		: std_logic;

	signal bank_add		: std_logic_vector(5 downto 0);

-- P6 I/F
	signal memcsn		: std_logic;
	signal piocsn		: std_logic;
	signal dobank		: std_logic_vector(7 downto 0);
	signal dop6if		: std_logic_vector(7 downto 0);

-- AVR I/F
-- ROMemu control
	signal pa_o			: std_logic_vector(7 downto 0);
	signal pa_cntn		: std_logic;
	signal romwrmoden	: std_logic;
	signal romwrn		: std_logic;
	signal adavr		: std_logic_vector(18 downto 0);
	signal doavr		: std_logic_vector(7 downto 0);

	signal r_noe_i		: std_logic;

	signal p6pio_pao	: std_logic_vector(7 downto 0);
	signal p6pio_pbo	: std_logic_vector(7 downto 0);
	signal p6pio_pco	: std_logic_vector(3 downto 0);
	signal avrpio_pao	: std_logic_vector(7 downto 0);
	signal avrpio_pbo	: std_logic_vector(7 downto 0);
	signal avrpio_pco	: std_logic_vector(3 downto 0);
	signal avrrstn		: std_logic;
	signal romsel		: std_logic;


	component P6IF is
		port (
			ADD			: in  std_logic_vector(15 downto 0);
			DTI			: in  std_logic_vector(7 downto 0);
			IORQN		: in  std_logic;
			WRN			: in  std_logic;
			RDN			: in  std_logic;
			MEMCSN		: in  std_logic;
			PIOCSN		: in  std_logic;
			ROMSEL		: in  std_logic;
			PAI			: in  std_logic_vector(7 downto 0);
			PBI			: in  std_logic_vector(7 downto 0);
			PCI			: in  std_logic_vector(3 downto 0);
			CLK			: in  std_logic;
			AVRRSTN		: in  std_logic;
			RSTN		: in  std_logic;
			BANK0		: out std_logic_vector(5 downto 0);
			BANK1		: out std_logic_vector(5 downto 0);
			BANK2		: out std_logic_vector(5 downto 0);
			BANK3		: out std_logic_vector(5 downto 0);
			BANK4		: out std_logic_vector(5 downto 0);
			BANK5		: out std_logic_vector(5 downto 0);
			BANK6		: out std_logic_vector(5 downto 0);
			BANK7		: out std_logic_vector(5 downto 0);
			ROMSET		: out std_logic_vector(1 downto 0);
			RAMSET		: out std_logic_vector(1 downto 0);
			CS01R		: out std_logic;
			CS01RX		: out std_logic;
			CS01W		: out std_logic;
			CS02W		: out std_logic;
			PC60MODE	: out std_logic;
			ROMXMODE	: out std_logic;
			CSCG		: out std_logic;
			DOBANK		: out std_logic_vector(7 downto 0);
			DO			: out std_logic_vector(7 downto 0);
			PAO			: out std_logic_vector(7 downto 0);
			PBO			: out std_logic_vector(7 downto 0);
			PCO			: out std_logic_vector(3 downto 0)
		);
	end component;

	component AVRIF is
		port (
			ADD			: in  std_logic_vector(2 downto 0);
			DTI			: in  std_logic_vector(7 downto 0);
			PAI			: in  std_logic_vector(7 downto 0);
			PBI			: in  std_logic_vector(7 downto 0);
			PCI			: in  std_logic_vector(3 downto 0);
			WRMODEN		: in  std_logic;
			DTOUTN		: in  std_logic;
			DTWRN		: in  std_logic;
			A_WRN		: in  std_logic;
			RSTN		: in  std_logic;
			PAO			: out std_logic_vector(7 downto 0);
			PBO			: out std_logic_vector(7 downto 0);
			PCO			: out std_logic_vector(3 downto 0);
			DTO			: out std_logic_vector(7 downto 0);
			AVRRSTN		: out std_logic;
			ROMSEL		: out std_logic;
			ROMAD		: out std_logic_vector(18 downto 0);
			ROMDT		: out std_logic_vector(7 downto 0);
			ROMWRN		: out std_logic;
			ROMWRMODEN	: out std_logic;
			DTOCNTN		: out std_logic;
			P6_NRSTOUT	: out std_logic
		);
	end component;

begin

-- P6Ƃ̃C^tF[X
	U_P6IF : P6IF
	port map (
		ADD			=> C_A,
		DTI			=> C_D,
		IORQN		=> C_NIORQ,
		WRN			=> C_NWR,
		RDN			=> C_NRD,
		MEMCSN		=> memcsn,
		PIOCSN		=> piocsn,
		ROMSEL		=> romsel,
		PAI			=> avrpio_pbo,
		PBI			=> avrpio_pao,
		PCI			=> avrpio_pco,
		CLK			=> C_4MCLK,
		AVRRSTN		=> avrrstn,
		RSTN		=> C_NRST,
		BANK0		=> bankset(0),
		BANK1		=> bankset(1),
		BANK2		=> bankset(2),
		BANK3		=> bankset(3),
		BANK4		=> bankset(4),
		BANK5		=> bankset(5),
		BANK6		=> bankset(6),
		BANK7		=> bankset(7),
		ROMSET		=> romset,
		RAMSET		=> ramset,
		CS01R		=> cs01r,
		CS01RX		=> cs01rx,
		CS01W		=> cs01w,
		CS02W		=> cs02w,
		PC60MODE	=> pc60mode,
		ROMXMODE	=> romxmode,
		CSCG		=> cscg,
		DOBANK		=> dobank,
		DO			=> dop6if,
		PAO			=> p6pio_pao,
		PBO			=> p6pio_pbo,
		PCO			=> p6pio_pco
	);


-- SW1,2
-- OFF(1),OFF(1)F66/66SR
-- OFF(1), ON(0)Fmk2/mk2SR/60
--  ON(0),OFF(1)FFDDsgp/փ|[ggpimk2/mk2SRpj
--  ON(0), ON(0)Fmk2/mk2SR/60
	piocsn <= 
		'0' when (SW(2 downto 1) = "11" and C_NOCS5 = '0') else
		'0' when (SW(2 downto 1) = "01" and C_A(7 downto 4) = "1101") else
		'0' when (SW(2 downto 1) = "10" and C_A(7 downto 4) = "0001") else
		'0' when (SW(2 downto 1) = "00" and C_A(7 downto 4) = "1101") else
		'1';

	memcsn <= '0' when (C_A(7 downto 4) = "0011") else '1';

-- AVRƂ̃C^tF[X
	U_AVRIF : AVRIF
	port map (
		ADD			=> PD(2 downto 0),
		DTI			=> PA,
		PAI			=> p6pio_pbo,
		PBI			=> p6pio_pao,
		PCI			=> p6pio_pco,
		WRMODEN		=> PD(4),
		DTOUTN		=> PD(5),
		DTWRN		=> PD(6),
		A_WRN		=> PD(3),
		RSTN		=> C_NRST,
		PAO			=> avrpio_pao,
		PBO			=> avrpio_pbo,
		PCO			=> avrpio_pco,
		DTO			=> pa_o,
		AVRRSTN		=> avrrstn,
		ROMSEL		=> romsel,
		ROMAD		=> adavr,
		ROMDT		=> doavr,
		ROMWRN		=> romwrn,
		ROMWRMODEN	=> romwrmoden,
		DTOCNTN		=> pa_cntn,
		P6_NRSTOUT	=> NRSTOUT
	);

	PA <= pa_o when (pa_cntn = '0') else (others => 'Z');


-- VDGORAMVRAMǍ(@̂)
	vdgread_n <= '0' when (C_NMREQ = '1' and C_NIORQ = '1' and C_NEXCAS = '0' and C_NRAS2 = '0') else '1';

-- Bank No.(VDGANZXl)
-- VDGANZXAAreaE/F(Bank4/5)o
	bank_add <=
		bankset( conv_integer( "10" & C_A(13 downto 13) ) ) when (vdgread_n = '0') else
																	-- VDGANZXBankŒ
		bankset( conv_integer( C_A(15 downto 13) ) );

-- ROM/RAM̃AhXo
	R_A(18 downto 17) <=
		adavr(18 downto 17) when (romwrmoden = '0') else
		romset when (bank_add(5 downto 4) = "00") else
		ramset when (bank_add(5 downto 4) = "01") else
		"11";
	R_A(16 downto 13) <=
		adavr(16 downto 13) when (romwrmoden = '0') else
		bank_add(3 downto 0);

	C_A(15 downto 13) <= (others => 'Z'); 
	C_A(12 downto  0) <=
		adavr(12 downto 0) when (romwrmoden = '0') else
		(others => 'Z');

-- AhX^f[^
	NADOUTENB <= '1' when (romwrmoden = '0') else '0';

	NDOUTCNT <=
		'1' when (romwrmoden = '0') else
--		'0' when (memcsn = '0' and C_NIORQ = '0' and C_NRD = '0') else
		'0' when (piocsn = '0' and C_NIORQ = '0' and C_NRD = '0') else
		r_noe_i;

	C_D <=
		doavr  when (romwrmoden = '0') else
--		dobank when (memcsn = '0' and C_NIORQ = '0' and C_NRD = '0') else
		dop6if when (piocsn = '0' and C_NIORQ = '0' and C_NRD = '0') else
		(others => 'Z');

-- PC-6001@AreaA,BRead/WriteɓROM𖳌ɂ邽߂̃tO
-- mk2ȍ~łPortF0HAreaA,BOROM/RAMɐݒ肵Ă_CS2,_CS3,_DRDo͂邪
-- @ɂPortF0HȂĂȂ̂_CS2,_CS3,_DRDo͂Ȃ
	read_ab_n  <= '0' when ( C_A(15 downto 14) = "00" and cs01r = '1' and pc60mode = '1') else '1';
	write_ab_n <= '0' when ( C_A(15 downto 14) = "00" and cs01w = '1' and pc60mode = '1') else '1';
	romemu_n   <= '0' when ( C_A(15 downto 14) = "00" and cs01rx = '1' and romxmode = '1' and cscg = '0') else '1';

-- PC-6001@AreaC,DRead/WritẽtO(ReadCS2,CS3Ɠ)
	write_cd_n <= '0' when ( C_A(15 downto 14) = "01" and cs02w = '1' and pc60mode = '1') else '1';

-- ROMtO
	C_NEROM <=
		'0' when (C_NMREQ = '0' and C_NRD = '0' and read_ab_n = '0') else
		'0' when (C_NMREQ = '0' and C_NRD = '0' and romemu_n = '0') else
		'1';

-- P6݁igpj
	C_NEXINT <= '1';

-- OROM/RAM̈P6ǂݏo

	r_noe_i <=
		'0' when (vdgread_n= '0') else										-- gROMVDGANZX
		'0' when (C_NMREQ = '0' and C_NRD = '0' and C_NCS2 = '0') else		-- gROMǂݏo
		'0' when (C_NMREQ = '0' and C_NRD = '0' and C_NCS3 = '0') else		-- gROMǂݏo
		'0' when (C_NMREQ = '0' and C_NRD = '0' and C_NEXCAS = '0' and C_NRAS2 = '0') else
																			-- gRAMǂݏo
		'0' when (C_NMREQ = '0' and C_NRD = '0' and read_ab_n = '0') else	-- @AreaA/Bǂݏo
		'0' when (C_NMREQ = '0' and C_NRD = '0' and romemu_n = '0') else	-- ROM emu
		'1';

	R_NOE <=
		'1' when (romwrmoden = '0') else
		r_noe_i;

-- OROM/RAM̈ɏ
	ramwrn <=
		'0' when (bank_add(5 downto 4) = "01" and C_NMREQ = '0' and C_NWR = '0') else
		'0' when (bank_add(5 downto 4) = "10" and C_NMREQ = '0' and C_NWR = '0') else
		'1';

	R0_NWR <=
		romwrn when (romwrmoden = '0') else
		'1';

	R1_NWR <=
		'0' when (ramwrn = '0' and C_NEXCAS = '0' and C_NRAS2 = '0') else	-- gRAM
		'0' when (ramwrn = '0' and write_ab_n = '0') else					-- @AreaA/B
		'0' when (ramwrn = '0' and write_cd_n = '0') else					-- @AreaC/D
		'1';


-- ANZX𔻒(ROM/RAM/)
-- ROM/RAMso

	R0_NCS <=
		'0' when (romwrmoden = '0') else
		'0' when (bank_add(5 downto 4) = "00") else
		'1';

	R1_NCS <=
		'1' when (romwrmoden = '0') else
		'0' when (bank_add(5 downto 4) = "01") else
		'1';



end RTL;
